from collections import deque


#from .serial_comm import SerialCommunication

class Sensors:
    """
    A class to manage all received data from sensors. Singleton class.
    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        """
        Ensure that only one instance of the class is created (singleton pattern).
        """
        if cls._instance is None:
            cls._instance = super(Sensors, cls).__new__(cls)
            cls._instance._initialize(*args, **kwargs)
        return cls._instance

    def _initialize(self, stepper_ids=[1,2], u_ids=[1,2,3,4], b_ids=[1,2,3,4], imu_connected=True, u_median_filter_len=3):
        """
        Initialize the sensor data storage, steppers, and median_filter for ultrasonics configuration.

        Args (as specified from config file):
            u_ids ([int]): Integer array of ids of ultrasonic sensors intialized.
            b_ids ([int]): Integer array of ids of bumper switches to be initialized.
            imu_connected (bool): Flag indicating if IMU is connected.
            u_median_filter_len (int): Length of the median filter window for ultrasonic sensors.
        """
        if getattr(self, '_initialized', False):
            return
        
        self.stepper_ids = stepper_ids
        self.u_ids = u_ids
        self.b_ids = b_ids
        self.imu_connected = imu_connected
        
        self.heartbeat = False # Flag to indicate if heartbeat is received
        self.handshake = False # Flag to indicate if handshake is successful
        
        self.ups_poll_interval = 1 # Polling interval for UPS sensor in seconds
        self.ups_discharging = False

        # UPS intitialization
        try:
            from .MPU import MPU as UPS
            self.ups = UPS(addr=0x41)
        except Exception as e:
            self.ups = None
            print("UPS sensor not connected: ", e)
            
        if self.ups:
            self.ups_data = {'bus_voltage': 0.0, 'shunt_voltage': 0.0, 'current': 0.0, 'power': 0.0, 'percent': 0.0, 'discharging': False}
        else:
            self.ups_data = None
        
        
        self.u_median_filter_len = u_median_filter_len

        self.u_sonic_data = {f'u_{u_id}': 0.0 for u_id in u_ids}  # Ultrasonic sensor data
        self.u_median_filter = {f'u_{u_id}': deque(maxlen=u_median_filter_len) for u_id in u_ids}  # Deques for u_sonic data median filter
        
        # IMU sensor data
        if imu_connected:
            self.imu_data = {'accel_x': 0.0, 'accel_y': 0.0, 'accel_z': 0.0, 'gyro_x': 0.0, 'gyro_y': 0.0, 'gyro_z': 0.0}

        self.b_switch_data = {f'b_{b_id}': False for b_id in b_ids}  # Bumper switch data

        # Stepper motor counts
        if stepper_ids:
            self.stepper_count = {}
            for id in stepper_ids:
                if id == 1:
                    self.stepper_count['S_L'] = 0
                elif id == 2:
                    self.stepper_count['S_R'] = 0
                else:
                    self.stepper_count[f"S_{id}"] = 0
                        
        self._initialized = True
        

    def request_sensor_data(self, sensor_type="u"):
        """
        Request sensor data from the robot.

        Args:
            sensor_type (str): The type of sensor data to request.
            "u" : ultrasonic data stored in self.u_sonic_data
            "i" : IMU data stored in self.imu_data
            "b" : bumper switch data stored in self.b_switch_data
            "sc" : stepper count data stored in self.stepper_count
        """
        if sensor_type == "u":
            return self.u_sonic_data
        elif sensor_type == "i":
            return self.imu_data
        elif sensor_type == "b":
            return self.b_switch_data
        elif sensor_type == "sc":
            return self.stepper_count
        else:
            return "Unknown request. Please specify a valid sensor type (u, i, b, sc)."

    def parse_data(self, data):
        """
        Parse the received data and update the sensor data dictionaries.
        
        Args:
            data (str): A string containing sensor data in the format "<type> <id> <values>".
            data format: "<identifier> <id> <values>"
            <values> could be multiple data.
        """
        parts = data.split()
        identifier = parts[0]
        if identifier == "h":
            self._manage_heartbeat_data()
            return
        elif identifier == "handshake":
            self.handshake = True
            return
        elif identifier =="BP":
            self._manage_bumper_press(parts[1])
            return
        elif identifier not in ["u", "i", "b", "sc"]:
            print("PICO message:", (" ").join(parts))
            return
        else:  
            try:
                sensor_id = parts[1]
                sensor_data = parts[2:]
            except Exception as e:
                print("Error parsing sensor data:")
                print("PICO message:", (" ").join(parts))
                return
            
        if identifier == "u":
            self._manage_u_sonic_data(sensor_id, sensor_data[0])
        elif identifier == "i":
            self._manage_imu_data(sensor_data)
        elif identifier == "b":
            self._manage_b_switch_data(sensor_id, sensor_data[0])
        elif identifier == "sc":
            self._manage_stepper_count(sensor_id, sensor_data)
    
    def _manage_bumper_press(self, b_id):
        """
        Manage bumper press data.
        
        Args:
            b_id (str): The ID of the bumper switch.
        """
        print(f"Bumper {b_id} pressed")

    def _median_filter(self, u_id, u_value):
        """
        Implement a median filter for ultrasonic sensor data.
        
        Args:
            u_id (str): The ID of the ultrasonic sensor.
            u_value (float): The latest reading from the ultrasonic sensor.
        """
        u_value = round(float(u_value), 3)
        if u_value == -0.017:
            u_value = 260.0
        elif u_value == -0.032:
            self.u_sonic_data[f'u_{u_id}'] = "sensor disconnected"
            return

        self.u_median_filter[f'u_{u_id}'].append(u_value)

        if len(self.u_median_filter[f'u_{u_id}']) == self.u_median_filter_len:
            sorted_values = sorted(self.u_median_filter[f'u_{u_id}'])
            median_value = sorted_values[self.u_median_filter_len // 2]
            self.u_sonic_data[f'u_{u_id}'] = median_value
            # Reset the deque
            self.u_median_filter[f'u_{u_id}'].clear()

    def _manage_u_sonic_data(self, u_id, u_data):
        """
        Manage ultrasonic sensor data.
        
        Args:
            u_id (str): The ID of the ultrasonic sensor.
            u_data (list): The list containing the data from the ultrasonic sensor.
        """
        self._median_filter(u_id, u_data)

    def _manage_imu_data(self, i_data):
        """
        Manage IMU sensor data.
        
        Args:
            i_data (list): The list containing the data from the IMU sensor.
        """
        self.imu_data['accel_x'] = float(i_data[0])
        self.imu_data['accel_y'] = float(i_data[1])
        self.imu_data['accel_z'] = float(i_data[2])
        self.imu_data['gyro_x'] = float(i_data[3])
        self.imu_data['gyro_y'] = float(i_data[4])
        self.imu_data['gyro_z'] = float(i_data[5])

    def _manage_b_switch_data(self, b_id, b_data):
        """
        Manage bumper switch data.
        
        Args:
            b_id (str): The ID of the bumper switch.
            b_data (str): The state of the bumper switch (True/False).
        """
        b_state = b_data.lower() == 'true'
        self.b_switch_data[f'b_{b_id}'] = b_state
    
    def _manage_stepper_count(self, stepper_id, count_data):
        """
        Manage stepper motor count data.
        
        Args:
            stepper_id (str): The ID of the stepper motor.
            count_data (list): The list containing the step count data from the stepper motor.
        """
        count = int(count_data[0])
        if stepper_id == '1':
            self.stepper_count['S_L'] = count
        elif stepper_id == '2':
            self.stepper_count['S_R'] = count
        else:
            self.stepper_count[f"S_{stepper_id}"] = count
    
    def _manage_heartbeat_data(self):
        """
        Manage heartbeat data.
        """
        print("Heartbeat received")
        self.heartbeat = True
        pass

    def get_ups_data(self):
        """
        Get the UPS sensor data.
        
        Return a dictionary of the UPS data at time of request and updates the data.
        """
        if not self.ups:
            return None
        bus_voltage = self.ups.getBusVoltage_V()             # voltage on V- (load side)
        shunt_voltage = self.ups.getShuntVoltage_mV() / 1000 # voltage between V+ and V- across the shunt
        current = self.ups.getCurrent_mA()                   # current in mA
        power = self.ups.getPower_W()                        # power in W
        p = (bus_voltage - 9)/3.6*100
        if(p > 100):p = 100
        if(p < 0):p = 0
        
        if current < 0:
            self.ups_discharging = True
        else:
            self.ups_discharging = False
        
        self.ups_data['bus_voltage'] = bus_voltage
        self.ups_data['shunt_voltage'] = shunt_voltage
        self.ups_data['current'] = current
        self.ups_data['power'] = power
        self.ups_data['percent'] = p
        self.ups_data['discharging'] = self.ups_discharging
        
        return self.ups_data

    def delete_instance():
        """
        Delete the instance of the class.
        """
        Sensors._instance = None

# Example usage
# sensors1 = Sensors(u_count=4, b_count=4, imu_connected=True, u_median_filter_len=3)
# sensors2 = Sensors(u_count=6, b_count=6, imu_connected=False, u_median_filter_len=5)

# # Simulated sensor data
# sensors1.parse_data("u 1 230.24738")
# sensors1.parse_data("u 1 240.12345")
# sensors1.parse_data("u 1 250.56789")

# sensors1.parse_data("b 3 False")
# sensors1.parse_data("i 1 0.1 6.5 8.2 20.243 18.3 9.1")

# print(sensors1.b_switch_data)
# print(sensors1.imu_data)

# print(sensors1 is sensors2)  # True, both are the same instance



default_config = {
    "SENSORS": {
        "imu": {
            "id": 1,
            "enabled": False,
            "scl_pin": 1,
            "sda_pin": 0,
            "polling_rate": 50
        },
        "bumper_switches": {
            "polling_rate": 1,
            "sensors": [
                {"id": 1, "enabled": False, "pin": 10},
                {"id": 2, "enabled": False, "pin": 11},
                {"id": 3, "enabled": True, "pin": 12},
                {"id": 4, "enabled": True, "pin": 13}
            ]
        },
        "ultrasonic": {
            "polling_rate": 2,
            "sensors": [
                {"id": 1, "enabled": False, "trigger_pin": 14, "echo_pin": 15},
                {"id": 2, "enabled": False, "trigger_pin": 16, "echo_pin": 17},
                {"id": 3, "enabled": True, "trigger_pin": 18, "echo_pin": 19},
                {"id": 4, "enabled": True, "trigger_pin": 20, "echo_pin": 21}
            ]
        }
    },
    "stepper_motors": [
        {
            "id": 1,
            "enabled": True,
            "step_pin": 3,
            "dir_pin": 4,
            "enable_pin": 2,
            "np_id": 1,
            "np_start" : 0,
            "np_end" : 7,
            "acc_step_size": 50,
            "acc_timer_period": 10
        },
        {
            "id": 2,
            "enabled": True,
            "step_pin": 6,
            "dir_pin": 7,
            "enable_pin": 5,
            "np_id": 2,
            "np_start" : 0,
            "np_end" : 7,
            "led_pin": 9,
            "acc_step_size": 50,
            "acc_timer_period": 10
        }
    ],
    "default_emergency_behavior": False,
    "heartbeat": {
        "enabled": False,
        "timer": 10000
    },
    "neopixel_strips":[
        {
            "id": 1,
            "enabled": True,
            "pin": 8,
            "num_pixels": 10
        },
        {
            "id": 2,
            "enabled": True,
            "pin": 9,
            "num_pixels": 10
        }
    ],
    "robot_parameters": {
        "microstepping": 0.5,
        "wheel_radius": 0.045,
        "wheel_separation": 0.295,
        "ticks_per_rev_full_step": 0.5
    }
}




import smbus
import time

# Config Register (R/W)
_REG_CONFIG                 = 0x00
# SHUNT VOLTAGE REGISTER (R)
_REG_SHUNTVOLTAGE           = 0x01

# BUS VOLTAGE REGISTER (R)
_REG_BUSVOLTAGE             = 0x02

# POWER REGISTER (R)
_REG_POWER                  = 0x03

# CURRENT REGISTER (R)
_REG_CURRENT                = 0x04

# CALIBRATION REGISTER (R/W)
_REG_CALIBRATION            = 0x05

class BusVoltageRange:
    """Constants for ``bus_voltage_range``"""
    RANGE_16V               = 0x00      # set bus voltage range to 16V
    RANGE_32V               = 0x01      # set bus voltage range to 32V (default)

class Gain:
    """Constants for ``gain``"""
    DIV_1_40MV              = 0x00      # shunt prog. gain set to  1, 40 mV range
    DIV_2_80MV              = 0x01      # shunt prog. gain set to /2, 80 mV range
    DIV_4_160MV             = 0x02      # shunt prog. gain set to /4, 160 mV range
    DIV_8_320MV             = 0x03      # shunt prog. gain set to /8, 320 mV range

class ADCResolution:
    """Constants for ``bus_adc_resolution`` or ``shunt_adc_resolution``"""
    ADCRES_9BIT_1S          = 0x00      #  9bit,   1 sample,     84us
    ADCRES_10BIT_1S         = 0x01      # 10bit,   1 sample,    148us
    ADCRES_11BIT_1S         = 0x02      # 11 bit,  1 sample,    276us
    ADCRES_12BIT_1S         = 0x03      # 12 bit,  1 sample,    532us
    ADCRES_12BIT_2S         = 0x09      # 12 bit,  2 samples,  1.06ms
    ADCRES_12BIT_4S         = 0x0A      # 12 bit,  4 samples,  2.13ms
    ADCRES_12BIT_8S         = 0x0B      # 12bit,   8 samples,  4.26ms
    ADCRES_12BIT_16S        = 0x0C      # 12bit,  16 samples,  8.51ms
    ADCRES_12BIT_32S        = 0x0D      # 12bit,  32 samples, 17.02ms
    ADCRES_12BIT_64S        = 0x0E      # 12bit,  64 samples, 34.05ms
    ADCRES_12BIT_128S       = 0x0F      # 12bit, 128 samples, 68.10ms

class Mode:
    """Constants for ``mode``"""
    POWERDOW                = 0x00      # power down
    SVOLT_TRIGGERED         = 0x01      # shunt voltage triggered
    BVOLT_TRIGGERED         = 0x02      # bus voltage triggered
    SANDBVOLT_TRIGGERED     = 0x03      # shunt and bus voltage triggered
    ADCOFF                  = 0x04      # ADC off
    SVOLT_CONTINUOUS        = 0x05      # shunt voltage continuous
    BVOLT_CONTINUOUS        = 0x06      # bus voltage continuous
    SANDBVOLT_CONTINUOUS    = 0x07      # shunt and bus voltage continuous


class MPU:
    def __init__(self, i2c_bus=1, addr=0x40):
        self.bus = smbus.SMBus(i2c_bus);
        self.addr = addr

        # Set chip to known config values to start
        self._cal_value = 0
        self._current_lsb = 0
        self._power_lsb = 0
        self.set_calibration_32V_2A()

    def read(self,address):
        data = self.bus.read_i2c_block_data(self.addr, address, 2)
        return ((data[0] * 256 ) + data[1])

    def write(self,address,data):
        temp = [0,0]
        temp[1] = data & 0xFF
        temp[0] =(data & 0xFF00) >> 8
        self.bus.write_i2c_block_data(self.addr,address,temp)

    def set_calibration_32V_2A(self):
        """Configures to INA219 to be able to measure up to 32V and 2A of current. Counter
           overflow occurs at 3.2A.
           ..note :: These calculations assume a 0.1 shunt ohm resistor is present
        """
        # By default we use a pretty huge range for the input voltage,
        # which probably isn't the most appropriate choice for system
        # that don't use a lot of power.  But all of the calculations
        # are shown below if you want to change the settings.  You will
        # also need to change any relevant register settings, such as
        # setting the VBUS_MAX to 16V instead of 32V, etc.

        # VBUS_MAX = 32V             (Assumes 32V, can also be set to 16V)
        # VSHUNT_MAX = 0.32          (Assumes Gain 8, 320mV, can also be 0.16, 0.08, 0.04)
        # RSHUNT = 0.1               (Resistor value in ohms)

        # 1. Determine max possible current
        # MaxPossible_I = VSHUNT_MAX / RSHUNT
        # MaxPossible_I = 3.2A

        # 2. Determine max expected current
        # MaxExpected_I = 2.0A

        # 3. Calculate possible range of LSBs (Min = 15-bit, Max = 12-bit)
        # MinimumLSB = MaxExpected_I/32767
        # MinimumLSB = 0.000061              (61uA per bit)
        # MaximumLSB = MaxExpected_I/4096
        # MaximumLSB = 0,000488              (488uA per bit)

        # 4. Choose an LSB between the min and max values
        #    (Preferrably a roundish number close to MinLSB)
        # CurrentLSB = 0.0001 (100uA per bit)
        self._current_lsb = .1  # Current LSB = 100uA per bit

        # 5. Compute the calibration register
        # Cal = trunc (0.04096 / (Current_LSB * RSHUNT))
        # Cal = 4096 (0x1000)

        self._cal_value = 4096

        # 6. Calculate the power LSB
        # PowerLSB = 20 * CurrentLSB
        # PowerLSB = 0.002 (2mW per bit)
        self._power_lsb = .002  # Power LSB = 2mW per bit

        # 7. Compute the maximum current and shunt voltage values before overflow
        #
        # Max_Current = Current_LSB * 32767
        # Max_Current = 3.2767A before overflow
        #
        # If Max_Current > Max_Possible_I then
        #    Max_Current_Before_Overflow = MaxPossible_I
        # Else
        #    Max_Current_Before_Overflow = Max_Current
        # End If
        #
        # Max_ShuntVoltage = Max_Current_Before_Overflow * RSHUNT
        # Max_ShuntVoltage = 0.32V
        #
        # If Max_ShuntVoltage >= VSHUNT_MAX
        #    Max_ShuntVoltage_Before_Overflow = VSHUNT_MAX
        # Else
        #    Max_ShuntVoltage_Before_Overflow = Max_ShuntVoltage
        # End If

        # 8. Compute the Maximum Power
        # MaximumPower = Max_Current_Before_Overflow * VBUS_MAX
        # MaximumPower = 3.2 * 32V
        # MaximumPower = 102.4W

        # Set Calibration register to 'Cal' calculated above
        self.write(_REG_CALIBRATION,self._cal_value)

        # Set Config register to take into account the settings above
        self.bus_voltage_range = BusVoltageRange.RANGE_32V
        self.gain = Gain.DIV_8_320MV
        self.bus_adc_resolution = ADCResolution.ADCRES_12BIT_32S
        self.shunt_adc_resolution = ADCResolution.ADCRES_12BIT_32S
        self.mode = Mode.SANDBVOLT_CONTINUOUS
        self.config = self.bus_voltage_range << 13 | \
                      self.gain << 11 | \
                      self.bus_adc_resolution << 7 | \
                      self.shunt_adc_resolution << 3 | \
                      self.mode
        self.write(_REG_CONFIG,self.config)

    def getShuntVoltage_mV(self):
        self.write(_REG_CALIBRATION,self._cal_value)
        value = self.read(_REG_SHUNTVOLTAGE)
        if value > 32767:
            value -= 65535
        return value * 0.01

    def getBusVoltage_V(self):
        self.write(_REG_CALIBRATION,self._cal_value)
        self.read(_REG_BUSVOLTAGE)
        return (self.read(_REG_BUSVOLTAGE) >> 3) * 0.004

    def getCurrent_mA(self):
        value = self.read(_REG_CURRENT)
        if value > 32767:
            value -= 65535
        return value * self._current_lsb

    def getPower_W(self):
        self.write(_REG_CALIBRATION,self._cal_value)
        value = self.read(_REG_POWER)
        if value > 32767:
            value -= 65535
        return value * self._power_lsb
        
if __name__=='__main__':

    # Create an INA219 instance.
    ina219 = MPU(addr=0x41)
    while True:
        bus_voltage = ina219.getBusVoltage_V()             # voltage on V- (load side)
        shunt_voltage = ina219.getShuntVoltage_mV() / 1000 # voltage between V+ and V- across the shunt
        current = ina219.getCurrent_mA()                   # current in mA
        power = ina219.getPower_W()                        # power in W
        p = (bus_voltage - 9)/3.6*100
        if(p > 100):p = 100
        if(p < 0):p = 0

        # INA219 measure bus voltage on the load side. So PSU voltage = bus_voltage + shunt_voltage
        #print("PSU Voltage:   {:6.3f} V".format(bus_voltage + shunt_voltage))
        #print("Shunt Voltage: {:9.6f} V".format(shunt_voltage))
        print("Load Voltage:  {:6.3f} V".format(bus_voltage))
        print("Current:       {:9.6f} A".format(current/1000))
        print("Power:         {:6.3f} W".format(power))
        print("Percent:       {:3.1f}%".format(p))
        print("")

        time.sleep(2)



from .midibot import DifferentialDriveRobot
from .serial_comm import SerialCommunication
from .socket_comm import SocketCommunication
from .sensors import Sensors

__all__ = ["DifferentialDriveRobot", "SerialCommunication", "SocketCommunication", "Sensors"]



import threading
import serial
from .sensors import Sensors
import time
import errno
import logging

class SerialCommunication:
    """
    A singleton class to manage serial communication.
    Handles connection, disconnection, sending, and receiving data via a serial port.
    """

    _instance = None
    _lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super(SerialCommunication, cls).__new__(cls)
                    cls._instance._initialized = False
        return cls._instance

    def __init__(self, serial_port, baudrate=115200, timeout=1):
        if self._initialized:
            return
        self.possible_ports = []
        self.serial_port = serial_port
        self.baudrate = baudrate
        self.timeout = timeout
        self.serial_connection = None
        self.running = False
        self.polling_thread = None
        self._internal_lock = threading.Lock()
        self._initialized = True
        self.data_callback = None

        self.sensors = Sensors()  # Access singleton instance of Sensors class
        self.ups_timer = time.time()

    def connect(self):
        """
        Establish a serial connection to the specified port and baudrate.
        """
        try:
            with self._internal_lock:
                if not self.serial_connection or not self.serial_connection.is_open:
                    self.serial_connection = serial.Serial(
                        self.serial_port, self.baudrate, timeout=self.timeout
                    )
                    self.running = True
                    if self.polling_thread and self.polling_thread.is_alive():
                        return
                    self.polling_thread = threading.Thread(target=self._poll_serial)
                    self.polling_thread.start()
                    print("Serial connection established.")
        except Exception as e:
            print("Serial connection error: ", e)
                
    def detect_ports(self):
        """
        Detect available serial ports.
        """
        for port in serial.tools.list_ports.comports():
            self.possible_ports.append(port.device)

    def disconnect(self):
        """
        Close the serial connection and stop the polling thread.
        """
        with self._internal_lock:
            self.running = False
            if self.polling_thread:
                self.polling_thread.join()
            if self.serial_connection:
                self.serial_connection.close()

    def _poll_serial(self):
        """
        Poll the serial port for incoming data in a separate thread.
        """
        while self.running:
            try:
                if self.serial_connection.in_waiting > 0:
                    with self._internal_lock:
                        raw_data = self.serial_connection.readline().decode("utf-8").strip()
                    if raw_data:
                        if self.data_callback:
                            self.data_callback(raw_data)
                        self.sensors.parse_data(raw_data)
                        if self.sensors.heartbeat:
                            self.send_command("h")
                            self.sensors.heartbeat = False
            except IOError as e:
                if e.errno == errno.EIO:
                    logging.error("Serial connection lost. Attempting to reconnect...")
                    if self.serial_connection:
                        try:
                            self.serial_connection.close()
                        except Exception as e:
                            logging.error(f"Error closing serial connection: {e}")
                        self.serial_connection = None

                    for attempt in range(1, 4):  # Attempt to reconnect 3 times
                        try:
                            self.connect()
                            if self.serial_connection:
                                logging.info("Reconnected to serial port successfully.")
                                break
                        except Exception as e:
                            logging.error(f"Reconnection attempt {attempt} failed: {e}")
                        
                        time.sleep(2 ** attempt)  # Exponential backoff strategy
                    else:
                        logging.critical("Failed to reconnect to serial port after multiple attempts.")
                        self.polling_thread.join()
                        self.__del__()  # Consider a more graceful shutdown approach
            
            if time.time() - self.ups_timer > self.sensors.ups_poll_interval:
                self.sensors.get_ups_data()
                self.ups_timer = time.time()
                
    
    def set_data_callback(self, callback):
        """
        Set a callback function to handle incoming data.
        """
        self.data_callback = callback

    def send_command(self, command):
        """
        Send a command through the serial connection.
        """
        with self._internal_lock:
            if self.serial_connection:
                self.serial_connection.write((command + "\n").encode("utf-8"))

    def is_connected(self):
        """
        Check if the serial connection is currently open.
        """
        with self._internal_lock:
            return self.serial_connection.is_open if self.serial_connection else False

    def __del__(self):
        self.disconnect()



import threading
import socket
from .sensors import Sensors


class SocketCommunication:
    """
    A singleton class to manage socket communication.
    Handles connection, disconnection, sending, and receiving data via a socket.
    """

    _instance = None
    _lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super(SocketCommunication, cls).__new__(cls)
                    cls._instance._initialized = False
        return cls._instance

    def __init__(self, ip="192.168.137.28", socket_port=8080):
        if self._initialized:
            return
        self.ip = ip
        self.socket_port = socket_port
        self.connection = None
        self.running = False
        self.polling_thread = None
        self._internal_lock = threading.Lock()
        self._initialized = True

        self.sensors = Sensors()  # Access singleton instance of Sensors class

    def connect(self):
        """
        Establish a socket connection to the specified IP and port.
        """
        with self._internal_lock:
            if not self.connection:
                self.connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.connection.settimeout(3)
                try:
                    self.connection.connect((self.ip, self.port))
                    self.running = True
                    self.polling_thread = threading.Thread(target=self._poll_socket)
                    self.polling_thread.start()
                except socket.timeout:
                    print("Socket connection timed out")
                except Exception as e:
                    print("Socket connection error:", e)

    def disconnect(self):
        """
        Close the socket connection and stop the polling thread.
        """
        with self._internal_lock:
            self.running = False
            if self.polling_thread:
                self.polling_thread.join()
            if self.connection:
                self.connection.close()
                self.connection = None

    def _poll_socket(self):
        """
        Poll the socket for incoming data in a separate thread.
        """
        while self.running:
            try:
                raw_data = self.receive_newline()
                if raw_data:
                    self.sensors.parse_data(raw_data)
            except Exception as e:
                print("Socket receive error:", e)

    def send_command(self, data):
        """
        Send a command through the socket connection.
        """
        with self._internal_lock:
            if self.connection:
                try:
                    data += "\n"
                    self.connection.send(data.encode())
                except Exception as e:
                    print("Socket send error:", e)

    def receive_newline(self):
        """
        Receive data from the socket until a newline character is encountered.
        """
        data = ""
        while self.running:
            try:
                chunk = self.connection.recv(1).decode("utf-8")
                if chunk == "\n":
                    return data
                data += chunk
            except Exception as e:
                print("Socket receive error:", e)

    def is_connected(self):
        """
        Check if the socket is currently connected.
        """
        with self._internal_lock:
            return self.connection is not None

    def __del__(self):
        self.disconnect()



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as file:
    long_description = file.read()

setup(
    name="romer-midibot",
    version="1.1.3",
    packages=find_packages(),
    install_requires=[
        "pyserial",
    ],
    author="Omar Habib",
    author_email="omar1farouk@gmail.com",
    description="A Python API for Romer's MIDIbot",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
    ],
    python_requires=">=3.6",
    license="Apache License 2.0",
)



import threading
import json
import time
import math
from .serial_comm import SerialCommunication
from .socket_comm import SocketCommunication
from .sensors import Sensors
from .default_configs import default_config


class DifferentialDriveRobot:
    """
    Singleton class to manage communication and control of a differential drive robot.
    Supports both serial and socket communication.
    """

    _instance = None  # Class-level attribute to hold the single instance
    _lock = threading.Lock()  # Class-level lock to ensure thread safety

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:  # First check (without lock)
            with cls._lock:  # Acquire lock to enter critical section
                if cls._instance is None:  # Double-check (with lock)
                    cls._instance = super(DifferentialDriveRobot, cls).__new__(cls)
                    cls._instance._initialized = False
        return cls._instance

    def __init__(self, serial_port, baudrate=115200, timeout=1, ip="192.168.137.28", socket_port=8080, config_file=None, stepper_ids=[1,2], u_ids=[1,2,3,4], b_ids=[1,2,3,4], imu_connected=False, u_median_filter_len=3, default_emergency_behavior = False):
        """
        Initialize the robot with connection parameters and setup communication interfaces.

        Args:
            serial_port (str): Serial port on pc or raspberry pi for connecting to the robot.
            baudrate (int, optional): Baud rate for serial communication. Defaults to 115200.
            timeout (int, optional): Timeout in seconds for communication operations. Defaults to 1.
            ip (str, optional): IP address of pico for socket communication. Defaults to "192.168.137.28".
            socket_port (int, optional): Port for socket communication on pc raspberry pi. Defaults to 8080.
            stepper_ids ([int], optional): Integer array of stepper motor IDs. Defaults to [1, 2].
            u_ids ([int], optional): Integer array of ultrasonic sensors. Defaults to [1, 2, 3, 4].
            b_ids ([int], optional): Integer array of bumper sensors. Defaults to [1, 2, 3, 4].
            imu_connected (bool, optional): Whether IMU is connected. Defaults to False.
            u_median_filter_len (int, optional): Length of median filter window. Defaults to 3.
            default_emergency_behavior (bool, optional): Default behavior of bumper switches coded on pico. Defaults to True.
                        Change to False if custom behavior is desired. Bumper switches are be polled for data if enabled.
        """
        if self._initialized:
            return

        # Initialize connection parameters provided by user
        self.serial_port = serial_port
        self.baudrate = baudrate
        self.ip = ip
        self.socket_port = socket_port
        self.timeout = timeout
        self.config_file = config_file
        self.configs = None

        # Initialize flags for connection states
        self.serial_running = False
        self.socket_running = False

        # If not config file, initialize sensor as provided in the arguments else, read from config file
        if not config_file:
            self.stepper_ids = stepper_ids
            self.u_ids = u_ids
            self.b_ids = b_ids
            self.imu_connected = imu_connected
            self.default_emergency_behavior = default_emergency_behavior

            # Default robot parameters. If config file is specified, these will be read from config file
            self._microstepping = 0.5 # half step
            self._wheel_radius = 0.045 # meters
            self._wheel_separation = 0.295 # meters
            self._ticks_per_rev_full_step = 600 # ticks per revolution at full step
        else:
            self.u_ids, self.b_ids, self.imu_connected, self.stepper_ids, self.default_emergency_behavior = self.read_config_file(config_file)

        # Initialize communication and sensor objects
        self.sensors = Sensors(self.stepper_ids, self.u_ids, self.b_ids, self.imu_connected, u_median_filter_len)
        self.serial_comm = SerialCommunication(serial_port, baudrate, timeout)
        self.socket_comm = SocketCommunication(ip, socket_port)

        # Initialize internal thread lock
        self._internal_lock = threading.Lock()
        self._initialized = True
    
    # Robot parameters
    @property
    def microstepping(self):
        return self._microstepping

    @property
    def wheel_radius(self):
        return self._wheel_radius

    @property
    def wheel_separation(self):
        return self._wheel_separation

    @property
    def ticks_per_rev(self):
        """
        Ticks per revolution of the motor at the current microstepping level. 
        Full step is the default microstepping level. = 600
        Half step = 600/(1/2) = 1200
        Quarter step = 600/(1/4) = 2400 and so on
        """
        return self._ticks_per_rev_full_step / self._microstepping

    def connect(self, connection_type=None):
        """
        Establish a connection to the robot using either serial or socket communication.
        Tries to connect using serial first, and falls back to socket if unsuccessful.

        Args:
            connection_type (str, optional): Type of connection ('serial' or 'socket'). Defaults to None.
        """
        if connection_type == "serial":
            self._connect_serial()
        elif connection_type == "socket":
            self._connect_socket()
        else:
            for _ in range(3):
                self._connect_serial()
                if self.serial_running:
                    break
            else:
                if not self.serial_running:
                    self._connect_socket()        

    def _connect_serial(self):
        """
        Attempt to establish a serial connection.
        """
        
        self.serial_comm.connect()
        self.serial_running = self.serial_comm.is_connected()
        self.pico_config()


    def _connect_socket(self):
        """
        Attempt to establish a socket connection.
        """
        try:
            self.socket_comm.connect()
            self.socket_running = self.socket_comm.is_connected()
            self.pico_config()
        except Exception as e:
            print("Socket connection error:", e)

    def disconnect(self):
        """
        Disconnect both serial and socket connections.
        """
        if self.serial_running:
            self.serial_comm.disconnect()
            self.serial_running = False
        if self.socket_running:
            self.socket_comm.disconnect()
            self.socket_running = False
    
    def read_config_file(self, config_file):
            """
            Read a config.json file and return arrays of enabled sensor IDs and polling rates.

            Args:
                config_file (str): Path to the config.json file.

            Returns: 
                ultrasonic_ids (list): List of enabled ultrasonic sensor IDs.
                bumper_switch_ids (list): List of enabled bumper switch IDs.
                imu_enabled (bool): Flag indicating whether IMU is enabled.
                stepper_ids (list): List of enabled stepper motor IDs.
                default_emergency_behavior (bool): Flag indicating whether default bumper behavior is enabled.

            """
            try:
                with open(config_file, 'r') as f:
                    config_data = json.load(f)
                    self.configs = config_data
            except FileNotFoundError:
                print(f"Error: Config file '{config_file}' not found")
                return None
            except json.JSONDecodeError:
                print(f"Error: Unable to parse JSON from config file '{config_file}'")
                return None

            ultrasonic_ids = []
            bumper_switch_ids = []
            imu_enabled = False
            stepper_ids = []
            default_emergency_behavior = False

            # Read ultrasonic sensor IDs
            if "SENSORS" in config_data and "ultrasonic" in config_data["SENSORS"]:
                for sensor in config_data["SENSORS"]["ultrasonic"]["sensors"]:
                    if sensor.get("enabled", False):
                        ultrasonic_ids.append(sensor["id"])

            # Read bumper switch IDs
            if "SENSORS" in config_data and "bumper_switches" in config_data["SENSORS"]:
                for sensor in config_data["SENSORS"]["bumper_switches"]["sensors"]:
                    if sensor.get("enabled", False):
                        bumper_switch_ids.append(sensor["id"])

            # Read IMU enabled status
            if "SENSORS" in config_data and "imu" in config_data["SENSORS"]:
                imu_enabled = config_data["SENSORS"]["imu"].get("enabled", False)

            # Read stepper motor IDs
            if "stepper_motors" in config_data:
                for motor in config_data["stepper_motors"]:
                    if motor.get("enabled", False):
                        stepper_ids.append(motor["id"])

            # Read enable default bumper behavior flag
            default_emergency_behavior = config_data.get("default_emergency_behavior", False)

            # Read robot parameters
            if "robot_parameters" in config_data:
                robot_params = config_data["robot_parameters"]
                self._microstepping = robot_params.get("microstepping", 0.5)
                self._wheel_radius = robot_params.get("wheel_radius", 0.045)
                self._wheel_separation = robot_params.get("wheel_separation", 0.295)
                self._ticks_per_rev_full_step = robot_params.get("ticks_per_rev_full_step", 600)


            return ultrasonic_ids, bumper_switch_ids, imu_enabled, stepper_ids, default_emergency_behavior

    def pico_config(self):
        """
        Configure the Pico with the sensor and motor configuration.
        """

        #Establish handshake and send configuration to Pico
        self.send_command("HANDSHAKE")
        
        while not self.sensors.handshake:
            print('User waiting for handshake...')
            time.sleep(0.5)

        if not self.configs:
            config = default_config

            # Update the configuration based on user input
            if "SENSORS" in config:
                if "ultrasonic" in config["SENSORS"]:
                    for sensor in config["SENSORS"]["ultrasonic"]["sensors"]:
                        if sensor["id"] in self.u_ids:
                            sensor["enabled"] = True
                if "bumper_switches" in config["SENSORS"]:
                    for sensor in config["SENSORS"]["bumper_switches"]["sensors"]:
                        if sensor["id"] in self.b_ids:
                            sensor["enabled"] = True
                if "imu" in config["SENSORS"]:
                    config["SENSORS"]["imu"]["enabled"] = self.imu_connected

            if "stepper_motors" in config:
                for motor in config["stepper_motors"]:
                    if motor["id"] in self.stepper_ids:
                        motor["enabled"] = True

            config["default_emergency_behavior"] = self.default_emergency_behavior

            self.configs = config
            self.send_command(json.dumps(config))
        else:
            self.send_command(json.dumps(self.configs))

    def pico_reconfig(self, config_file):
        """
        Reconfigure the Pico with a new configuration file. First a pause command is sent to the Pico to stop all movement and sensor polling.
        No data will be received from the Pico until a continue command is sent. Then the new configuration file is read and if successful,
        a the old instance of Sensors is deleted and a new instance is created with the new configuration. Finally, a reconfig command is sent
        to the Pico to apply the new configuration.
        """
        self.send_pause_command()
        self.config_file = config_file
        self.u_ids, self.b_ids, self.imu_connected, self.stepper_ids, self.default_emergency_behavior = self.read_config_file(config_file)
        self.sensors.delete_instance()
        self.sensors = Sensors(self.stepper_ids, self.u_ids, self.b_ids, self.imu_connected, self.u_median_filter_len)
        self.send_reconfig_command()
        time.sleep(1)
        self.pico_config()
        
    def set_data_callback(self, callback):
        """
        Set a callback function to handle incoming data from the robot.

        Args:
            callback (function): Callback function to handle incoming data.
        """
        if self.serial_running:
            self.serial_comm.set_data_callback(callback)
        elif self.socket_running:
            self.socket_comm.set_data_callback(callback)

    def send_command(self, command):
        """
        Send a command to the robot using either serial or socket communication.

        Args:
            command (str): Command to be sent to the robot.
        """
        if self.serial_running:
            self.serial_comm.send_command(command)
        if self.socket_running:
            self.socket_comm.send_command(command)
    
    def send_pause_command(self):
        """
        Send a pause command to the robot. This will pause the robot's movement and sensor polling.
        Robot will be waiting for a continue command to resume. Previous robot movement commands will not be stored.
        All commands other than continue will be ignored.
        """
        self.send_command("PAUSE")
    
    def send_continue_command(self):
        """
        Send a continue command to the robot. This will resume the robot's sensor polling.
        """
        self.send_command("CONTINUE")
    
    def send_reconfig_command(self):
        """
        Send a reconfig command to the robot. This will reconfigure the robot with the new configuration file.
        """
        self.send_command("re-config")


    def set_ticks_duration(self, left_ticks, right_ticks, duration_l, duration_r):
        """
        Set the number of ticks and duration for the left and right wheels of the robot.

        Args:
            left_ticks (int): Number of ticks for the left wheel.
            right_ticks (int): Number of ticks for the right wheel.
            duration_l (float): Duration to reach the left ticks in seconds.
            duration_r (float): Duration to reach the right ticks in seconds.
        """
        command_l = f"t l {left_ticks} {duration_l}"
        command_r = f"t r {right_ticks} {duration_r}"
        self.send_command(command_l)
        self.send_command(command_r)
    
    def set_speed(self, left_speed, right_speed):
        """
        Set the speed of the left and right wheels of the robot.

        Args:
            left_speed (float): Speed of the left wheel. 
            right_speed (float): Speed of the right wheel.
        """
        command_l = f"s l {left_speed}"
        command_r = f"s r {right_speed}"
        self.send_command(command_l)
        self.send_command(command_r)

    def move_forward(self, duration=1, speed=500):
        """
        Move the robot forward for a specified duration.

        Args:
            duration (float): Duration to move forward in seconds.
            speed (int): Speed at which to move forward. Defaults to 500.
                                minimum = 100 (ideally min=250), maximum = 4000 (ideally max=3600)
        """
        self.set_speed(speed, speed)
        time.sleep(duration)
        self.stop()
    
    def rotate(self, duration=5, speed=600, direction="cw"):
        """
        Rotate the robot at a given speed in a give direction for a set duration.

        Args:
            duration (float): Duration of rotation in seconds. Defaults to 5 s
            speed (int): Speed of rotation in frequency of wheel rotation. Defaults to 600 Hz
            direction (string): 'cw' for clockwise roation. 'ccw' for counter-clockwise rotation
        
        """
        if direction.lower() == "cw":
            self.set_speed(speed, -speed)
        elif direction.lower() == "ccw":
            self.set_speed(-speed, speed)
        else:
            print("Set direction is incorrect. direction should be 'cw' or 'ccw'.")
            return
        time.sleep(duration)
        self.stop()

    def send_twist(self, linear_speed, angular_speed):
        """
        Convert desired linear and angular speeds to wheel frequencies and set the robot speed.

        Args:
            linear_speed (float): Desired linear speed in meters per second.
            angular_speed (float): Desired angular speed in radians per second.
        """
        w_l = (linear_speed - (angular_speed * self._wheel_separation / 2)) / self._wheel_radius
        w_r = (linear_speed + (angular_speed * self._wheel_separation / 2)) / self._wheel_radius

        f_l = w_l / (2 * math.pi) * self._ticks_per_rev
        f_r = w_r / (2 * math.pi) * self._ticks_per_rev

        f_l = round(f_l, 3)
        f_r = round(f_r, 3)

        self.set_speed(f_l, f_r)

    def stop(self):
        """
        Stop the robot by setting both wheel speeds to zero.
        """
        self.set_speed(0, 0)
    
    def emergency_stop(self):
        """
        Stop the robot immediately by sending an emergency stop command.
        """
        self.send_command("EMERGENCY_STOP")
    
    def turn_leds_off(self):
        """
        Turn off the LEDs on the robot.
        """
        self.send_command("np off")
    
    def fill_leds(self, color):
        """
        Fill all LEDs with a given color.

        Args:
            color (tuple): RGB color tuple.
        """
        self.send_command(f"np fill {color[0]} {color[1]} {color[2]}")

    def set_led_pixel(self, np_id, pixel_index, color):
        """
        Set the color of a single LED pixel.

        Args:
            np_id (int): ID of the neopixel strip.
            pixel_index (int): Index of the pixel to set.
            color (tuple): RGB color tuple. Should be integer in the range [0, 255]
        """
        self.send_command(f"np set {np_id} {pixel_index} {color[0]} {color[1]} {color[2]}")

    def get_sensor_data(self, sensor_type="u"):
        """
        Get sensor data from the robot. Default is ultrasound sensor data.
        """
        return self.sensors.request_sensor_data(sensor_type)
    
    def get_status(self):
        """
        Request the current status of the robot. Two steps. Firs it prints connection status 
        and the sensors enabled. Then it sends a command to the robot to get the status.
        So it should print a nicely formatted status of the robot connection and connected sensors and steppers

        """
        print("Connection status:")
        if self.serial_running:
            print("Serial connection established")
        if self.socket_running:
            print("Socket connection established")
        print("Sensors enabled:")
        print("Ultrasonic sensors:", self.u_ids)
        print("Bumper switches:", self.b_ids)
        print("IMU enabled:", self.imu_connected)
        print("Stepper motors enabled:", self.stepper_ids)
        print("Default emergency behavior enabled:", self.default_emergency_behavior)
        self.send_command("GET_STATUS")
    
    def __del__(self):
        """
        Destructor to stop the robot and disconnect on deletion.
        """
        self.stop()
        self.disconnect()

